/**************************************************************************
 * SPICE2XGRAPH.C 							  *
 * 									  *
 * This program converts spice output into a format suitable for reading  *
 * into xgraph.  							  *
 * 									  *
 * usage: spice2xgraph < spice.out > xgraph.in 				  *
 *    or: spice < spice.in | spice1xgraph | xgraph 			  *
 *    or: spice2xgraph spice.out xgraph.in 				  *
 * 									  *
 * 									  *
 * Copyright (c) 1993, 1994 David Frey                                    * 
 * 									  *
 * This program is free software; you can redistribute it and/or modify   *
 * it under the terms of the GNU General Public License as published by   *
 * the Free Software Foundation; either version 2 of the License, or      *
 * (at your option) any later version.                                    *
 *									  * 
 * This program is distributed in the hope that it will be useful,        *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of         *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 * GNU General Public License for more details.                           *
 *									  *   
 * You should have received a copy of the GNU General Public License      *
 * along with this program; if not, write to the Free Software            *
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.              *
 *									  * 
 **************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <getopt.h>
#include "sp2xg.h"

static char RCSId[] = "sp2xg.c,v 1.0 1994/02/21 09:59:13 david Exp";

char title[MAXLINELENGTH];			 /* circuit name	      */
char XTitle[MAXLINELENGTH];			 /* label of X-Axis    */
char YTitle[8][MAXLINELENGTH];			 /* label(s) of Y-Axis */
int  NrOfGraphs;				 /* how many graphs we have */

struct XDataPointType *StartX;
struct YDataPointType *StartY[8];

char *progname;
int verbose;                                     /* be verbose        */

int extract, n;

char line[MAXLINELENGTH];			 /* current line read */
char *ok;

void usage(void)
{
  fprintf(stderr, "usage: %s [-v] [-t[n]|-a[n]] [-h] [-?] [-i spice.in] [-o xgraph.out]\n", progname);
  fprintf(stderr, "       %s [-v] [-t[n]|-a[n]] [-h] [-?] < spice.in > xgraph.out\n", progname);
  fprintf(stderr, "       spice|%s [-v] [-t[n]|-a[n]]|xgraph\n", progname);
  fprintf(stderr, "\n");
  fprintf(stderr, "  -v    --verbose        : verbose output\n");
  fprintf(stderr, "  -t[n] --transient [n]  : extract `transient' table (generated by .PRINT TRAN)\n");
  fprintf(stderr, "                           (the parameter n extracts the nth occurrence)\n");
  fprintf(stderr, "  -a[n] --ac [n]         : extract `ac' table (generated by .PRINT AC)\n");
  fprintf(stderr, "                           (the parameter n extracts the nth occurrence)\n");
  fprintf(stderr, "  -h    --help           : this help screen.\n");
  fprintf(stderr, "\n");
  fprintf(stderr, "  -i fn --input  filename: take filename as input\n");
  fprintf(stderr, "  -o fn --output filename: write output to filename\n");   
  fprintf(stderr, "\n");
  fprintf(stderr, "if neither the -t nor the -a switch have been supplied, the \n");
  fprintf(stderr, "first of the two tables will be chosen.\n");
}

void PrintError(int errmsgno, const char param[])
{
  char errmsg[80];

  switch (errmsgno)
  {
    case FETCHLINE:   strcpy (errmsg,"unable to fetch line!");                      break;
    case NODATA:      strcpy (errmsg,"No data to read!");                           break;
    case SIGNOTFOUND: sprintf(errmsg,"%s-analysis: signature not found!", param);   break;
    case NOTFOUND:    sprintf(errmsg, "%s not found (not open for input).", param); break;
    case DISKFULL:    sprintf(errmsg, "unable to create %s (disk full?)\n", param); break;
    case GARBAGE:     strcpy(errmsg, "Read garbage instead of data!");              break;
  }
  fprintf(stderr, "%s: %s\n", progname, errmsg);
  exit(1);
}

void RemoveDataPoints(void)
{
  struct XDataPointType *px;
  struct YDataPointType *py;
  struct XDataPointType *prevx;
  struct YDataPointType *prevy;
  struct XDataPointType *prevprevx;
  struct YDataPointType *prevprevy;
  int i;
 
  if (verbose) fprintf(stderr, "Removing heap ...");
   
  prevprevx = NULL;  
  while (StartX->nextX != NULL)
  {
    px = StartX; prevx = StartX;
    while (px->nextX != NULL)
    {
      prevprevx = prevx; prevx = px; px = px->nextX;
    }
    free(prevx);
    prevprevx->nextX = NULL;
  }
  free(StartX);

  if (verbose) fprintf(stderr, "X,");

  prevprevy = NULL;
  for (i = 0; i < NrOfGraphs; i++)
  {
    while (StartY[i]->nextY != NULL)
    {
      py = StartY[i]; prevy = StartY[i];
      while (py->nextY != NULL)
      {
	prevprevy = prevy; prevy = py; py = py->nextY;
      }
      free(prevy);
      prevprevy->nextY = NULL;
    }
    free(StartY[i]);
  }
  if (verbose) fprintf(stderr, "Y, done\n");
}

void ReadAndStoreTitles(FILE * infile, FILE * outfile)
{
  char dummy[MAXLINELENGTH];			 /* a dummy line       */
  char analysis[9];				 /* type of analysis   */
  int i;

  rewind(infile);

  /* The title can be found on line 3 (no error checking done) */

  for (i = 1; i < 3; ++i)
    ok = fgets(dummy, MAXLINELENGTH, infile);

  if (ok == NULL) PrintError(FETCHLINE, "");

  ok = fgets(line, MAXLINELENGTH, infile);
  if (ok == NULL) PrintError(FETCHLINE, "");

  ok = strtok(line, " ");
  fprintf(outfile, "TitleText: %s\n", strtok(NULL, "\n"));

  /* Search for template-line no n */
  for (i = 1; i <= n; i++)
  {
    strcpy(line, "");
    while (!(feof(infile) || ((strncmp(line, "0**** ", 6) == 0) &&
	    (((extract == FIRST) && ((strstr(line, template[AC]) != NULL) ||
				 (strstr(line, template[TRAN]) != NULL))) ||
       ((extract != FIRST) && (strstr(line, template[extract]) != NULL))))))
      ok = fgets(line, MAXLINELENGTH, infile);
  }

  switch (extract)
  {
    case FIRST: strcpy(analysis, "first");     break;
    case AC:    strcpy(analysis, "AC");        break;
    case TRAN:  strcpy(analysis, "transient"); break;
  }

  if (feof(infile)) PrintError(SIGNOTFOUND, template[extract]);
  if (ok == NULL)   PrintError(FETCHLINE, "");

  do
  {
    ok = fgets(dummy, MAXLINELENGTH, infile);
    if (ok == NULL) PrintError(FETCHLINE, "");

    if (dummy[0] != 'x') strcpy(line, dummy);
  }
  while (!feof(infile) && (dummy[0] != 'x'))
    ;

  if (feof(infile)) PrintError(SIGNOTFOUND, template[extract]);

  NrOfGraphs = sscanf(line, "%s %s %s %s %s %s %s %s %s",
		      &XTitle,
		      &YTitle[0], &YTitle[1], &YTitle[2], &YTitle[3],
		      &YTitle[4], &YTitle[5], &YTitle[6], &YTitle[7]);
  NrOfGraphs--;					 /* XTitle doesn't count */
  
  if (verbose) fprintf(stderr, "Number of graphs to display: %i\n", NrOfGraphs);
  if (NrOfGraphs <= 0) PrintError(NODATA, "");

  if (strstr(XTitle, "time") != NULL) fprintf(outfile, "XUnitText: t\n");
  else
  {
    if (strstr(XTitle, "freq") != NULL)
    {
      fprintf(outfile, "XUnitText: f\n");
      fprintf(outfile, "LogX\n");
    }
    else
      fprintf(outfile, "XUnitText: %s\n", XTitle);
  }

  /* next line is blank */
  ok = fgets(dummy, MAXLINELENGTH, infile);
  if (ok == NULL) PrintError(FETCHLINE, "");
}

void ReadAndStoreDataPoints(FILE * infile, FILE * outfile)
{
  int i;
  int eod;					 /* end of data */
  int valok;

  struct XDataPointType *LastX;
  struct YDataPointType *LastY[8];
  struct XDataPointType *NextX;
  struct YDataPointType *NextY[8];
  struct XDataPointType *PrevX;
  struct YDataPointType *PrevY[8];

  StartX = (struct XDataPointType *) malloc(sizeof(struct XDataPointType));

  for (i = 0; i < NrOfGraphs; i++)
    StartY[i] = (struct YDataPointType *) malloc(sizeof(struct YDataPointType));

  LastX = StartX;
  for (i = 0; i < NrOfGraphs; i++)
    LastY[i] = StartY[i];

  if (verbose) fprintf(stderr, "Reading data,");
  eod = FALSE;
  while (!feof(infile) && !eod)
  {
    ok = fgets(line, MAXLINELENGTH, infile);
    if (ok == NULL) PrintError(FETCHLINE, "");

    eod = (strchr(line, 'y') != NULL);
    if (!eod)
    {
      valok = sscanf(line, "%E %E %E %E %E %E %E %E %E",
		     &(LastX->X),
		     &(LastY[0]->Y), &(LastY[1]->Y), &(LastY[2]->Y), 
		     &(LastY[3]->Y), &(LastY[4]->Y), &(LastY[5]->Y), 
		     &(LastY[6]->Y), &(LastY[7]->Y));
      if (valok == 0)
      {
	RemoveDataPoints(); PrintError(GARBAGE, "");
      }
      NextX = (struct XDataPointType *) malloc(sizeof(struct XDataPointType));

      NextX->nextX = NULL;
      for (i = 0; i < NrOfGraphs; i++)
      {
	NextY[i] = (struct YDataPointType *) malloc(sizeof(struct YDataPointType));

	NextY[i]->nextY = NULL;
      }

      LastX->nextX = NextX;
      for (i = 0; i < NrOfGraphs; i++)
	LastY[i]->nextY = NextY[i];

      PrevX = LastX;
      LastX = NextX;
      for (i = 0; i < NrOfGraphs; i++)
      {
	PrevY[i] = LastY[i];
	LastY[i] = NextY[i];
      }
    }
  }
  if (verbose) fprintf(stderr, "done\n");
}

void PrintData(FILE * outfile)
{
  struct XDataPointType *px;
  struct YDataPointType *py;
  int i;

  if (verbose) fprintf(stderr, "Writing data in XGraph format");
  for (i = 0; i < NrOfGraphs; i++)
  {
    fprintf(outfile, "YUnitText: %s\n", YTitle[i]);
    fprintf(outfile, "\" %s\n", YTitle[i]);	 /* Set-description */

    px = StartX;
    py = StartY[i];
    while (px->nextX != NULL)
    {
      fprintf(outfile, "%20E %20E\n", px->X, py->Y);
      px = px->nextX;
      py = py->nextY;
    }
    fprintf(outfile, "\n");
  }
  if (verbose) fprintf(stderr, "done\n");
}

void ConvertSpiceInput(FILE * infile, FILE * outfile)
{
  ReadAndStoreTitles(infile, outfile);
  ReadAndStoreDataPoints(infile, outfile);
  PrintData(outfile);
  RemoveDataPoints();
}

int main(int argc, char *argv[])
{
  FILE *SpiceIn;
  FILE *XGraphOut;

  char c;

  progname = argv[0];

  fprintf(stderr,"%s version 1.0. Copyright (c) 1993, 1994 David Frey.\n",progname);
  fprintf(stderr,"%s is free software and you are welcome to distribute copies of it\n",progname);
  fprintf(stderr,"under certain conditions; you should have received a copy of these\n");
  fprintf(stderr,"conditions (see file 'COPYING') with this software.\n\n");
 
  ok = (char *) malloc(MAXLINELENGTH + 1);

  verbose = FALSE; extract   = FIRST;
  SpiceIn = stdin; XGraphOut = stdout;
  while ((c = getopt_long(argc, argv, "a:h?i:o:t:vV",
			  long_options, (int *) 0)) != EOF)
  {
    switch (c)
    {
      case 0  : break;
      case 'a':
                extract = AC;
                n = atoi(optarg);
                if (n < 1) n = 1;
                break;
      case 'h':
      case '?': usage(); return 0; break;
      case 'i':
                if ((SpiceIn   = fopen(optarg,"r")) == NULL) PrintError(NOTFOUND, optarg);
                break; 
      case 'o':
                if ((XGraphOut = fopen(optarg,"w")) == NULL) PrintError(DISKFULL, optarg);
                break;
      case 't':
                extract = TRAN;
                n = atoi(optarg);
                if (n < 1) n = 1;
                break;
      case 'v': verbose = TRUE; break;
      case 'V': fprintf(stdout,"%s\n",RCSId); return 0; break;
    }
  }

  if (verbose)
    if (extract == FIRST)
      fprintf(stderr, "Extracting first appearing table.\n");
    else
      fprintf(stderr, "Extracting %s table no %i\n.", template[extract], n);

  ConvertSpiceInput(SpiceIn, XGraphOut);

  fclose(SpiceIn); fclose(XGraphOut);

  free(ok);
  return 0;
}						 /* main */
